package bnrobot.clustering.hierarchical;

import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.PriorityQueue;
import java.util.TreeMap;

import bnrobot.configuration.CollectionConfiguration;
import bnrobot.core.Clustering;
import bnrobot.core.DocCollection;
import bnrobot.core.Similarity;
import bnrobot.core.TfIdfElement;
import bnrobot.core.models.DocumentElement;
import bnrobot.core.models.Subject;

/**
 * Implementação da HierarchicalClustering
 * 
 * @author willian
 * 
 */
public class HierarchicalClustering implements Clustering {
	private final DocCollection collection;
	private final CollectionConfiguration config;

	private final PriorityQueue<Similarity<Group, Group>> queue = new PriorityQueue<Similarity<Group, Group>>();

	private TfIdfElement<Group>[] groups;
	private List<TfIdfElement<Subject>> subjects;
	private TreeMap<Integer, Subject> output;

	public HierarchicalClustering(DocCollection collection,
			CollectionConfiguration config) {
		this.collection = collection;
		this.config = config;
	}

	/**
	 * Método auxiliar do Hierarchical Clustering Serve para criar um novo
	 * grupo,
	 * 
	 * @param p
	 */
	private void addSimilarities(int p) {
		for (int i = 0; i < p; i++) {
			if (this.groups[i] != null) {
				Similarity<Group, Group> similarity = new Similarity<Group, Group>(
						this.groups[p], this.groups[i]);
				if (similarity.getValue() >= this.config.getSubjectFloor()) {
					this.queue.add(similarity);
				}
			}
		}
	}

	/**
	 * Clusteriza os documentos: Hierarchical Clustering
	 * 
	 * @param docs
	 */
	private void clusterize(Collection<DocumentElement> docs) {
		this.groups = new TfIdfElement[2 * docs.size()];
		int i = 0;

		for (DocumentElement element : docs) {
			Group group = new Group(i, element);
			this.groups[i] = new TfIdfElement<Group>(this.collection.getIdf(),
					group);
			this.addSimilarities(i);
			i++;
		}

		while (!this.queue.isEmpty()) {
			Similarity<Group, Group> similarity = this.queue.poll();
			if (this.groups[similarity.getElementA().getId()] != null
					&& this.groups[similarity.getElementB().getId()] != null) {

				Group group = new Group(i, similarity.getElementA(),
						similarity.getElementB());
				this.groups[i] = new TfIdfElement<Group>(
						this.collection.getIdf(), group);
				this.addSimilarities(i);

				i++;
				this.groups[similarity.getElementA().getId()] = null;
				this.groups[similarity.getElementB().getId()] = null;
			}
		}
	}

	/**
	 * Dado uma lista de documentos, retorna os grupos de documentos formados
	 * pelo documentos mais similares
	 * 
	 * @param docs
	 * @return
	 */
	private List<Group> makeGroups(Collection<DocumentElement> docs) {
		this.clusterize(docs);

		List<Group> groups = new LinkedList<Group>();
		for (TfIdfElement<Group> grupo : this.groups) {
			if (grupo != null) {
				groups.add(grupo.getElement());
			}
		}
		this.groups = null;
		return groups;
	}

	/**
	 * Como os valores dos vetores mudam com o tempo (pois dependem da
	 * quantidade de documentos na coleção) antes de fazer os calculos eu faço
	 * um "cache" dos valores atuais de cada elemento, pois durante o
	 * processamento a quantidade de documentos é (ou pelo menos deveria) ser
	 * constante.
	 */
	private void cacheSubjects() {
		Iterable<Subject> subjects = this.collection.getActiveSubjects();

		List<TfIdfElement<Subject>> cached = new LinkedList<TfIdfElement<Subject>>();
		for (Subject subject : subjects) {
			cached.add(new TfIdfElement<Subject>(this.collection.getIdf(),
					subject));
		}
		this.subjects = cached;
	}

	/**
	 * Procura na lista de Subject ativos o subject com maior similaridade com o
	 * valor de entrada, que seja maior que a configurada.
	 * 
	 * @param group
	 *            Grupo de documentos
	 * @return Subject de maior similaridade, ou null se nenhum for maior que a
	 *         similaridade mínima.
	 * 
	 */
	private Subject findSubject(Group group) {
		if (this.subjects.size() == 0) {
			return null;
		}
		Iterator<TfIdfElement<Subject>> subjects = this.subjects.iterator();
		Similarity<Group, Subject> max = new Similarity<Group, Subject>(group,
				subjects.next(), this.collection);

		while (subjects.hasNext()) {
			Similarity<Group, Subject> sim = new Similarity<Group, Subject>(
					group, subjects.next(), this.collection);
			if (sim.getValue() > max.getValue()) {
				max = sim;
			}
		}

		return max.getValue() >= this.config.getFamilyFloor() ? max
				.getElementB() : null;
	}

	/**
	 * Faz atualização grupo de documents, ie ( seta o valor do subject em cada
	 * elemento e recalcula o ponto médio do grupo )
	 * 
	 * @param subject
	 *            Subject onde serão inseridos os documentos
	 * @param group
	 *            Grupo de documentos que serão inseridos no Subject
	 */
	private void updateSubject(Subject subject,
			Iterator<DocumentElement> elements) {
		while (elements.hasNext()) {
			DocumentElement element = elements.next();
			subject.addDocumentElement(element);
			element.setSubjectId(subject.getId());
		}
		if (!this.output.containsKey(subject.getId())) {
			this.output.put(subject.getId(), subject);
		}
	}

	private void createSubject(Group group) {
		Iterator<DocumentElement> elements = group.getElements().iterator();
		DocumentElement element = elements.next();

		Subject subject = this.collection.createSubject(element);
		this.updateSubject(subject, elements);

	}

	public synchronized Collection<Subject> insereDocs(
			Collection<DocumentElement> docs) {
		List<Group> groups = this.makeGroups(docs);
		this.cacheSubjects();

		this.output = new TreeMap<Integer, Subject>();

		for (Group group : groups) {
			Subject subject = this.findSubject(group);
			if (subject == null) {
				this.createSubject(group);
			} else {
				this.updateSubject(subject, group.getElements().iterator());
			}
		}

		return new LinkedList<Subject>(this.output.values());
	}

	public void commit() {
		// Não precisa commitar
	}

}
