package bnrobot.core;

import java.io.IOException;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;

import bnrobot.core.controller.CollectionController;
import bnrobot.core.models.DocumentElement;
import bnrobot.core.models.LabeledDocument;
import bnrobot.core.models.Subject;
import bnrobot.core.models.Term;
import bnrobot.core.provider.DocumentProvider;
import bnrobot.core.provider.FamilyProvider;
import bnrobot.core.provider.LabeledDocumentProvider;
import bnrobot.core.provider.SubjectProvider;
import bnrobot.core.provider.TermProvider;
import bnrobot.jdbm.persister.DocumentElementDao;
import bnrobot.jdbm.persister.FamilyDao;
import bnrobot.jdbm.persister.LabeledDocumentDao;
import bnrobot.jdbm.persister.SubjectDao;
import bnrobot.jdbm.persister.TermDao;
import bnrobot.jdbm.provider.FamilyCache;
import bnrobot.jdbm.provider.SubjectCache;
import bnrobot.jdbm.provider.TermCache;
import bnrobot.tests.Cronometro;

public class DocCollection {

	private final CollectionController controller;

	private final TermProvider termProvider;
	private final FamilyProvider familyProvider;
	private final SubjectProvider subjectProvider;
	private final DocumentProvider documentProvider;

	private final LabeledDocumentProvider labeledDocumentProvider;

	private Queue<DocumentElement> queue = new LinkedList<DocumentElement>();

	private double[] idf = null;

	public DocCollection(CollectionController controller) {
		this.controller = controller;

		this.termProvider = new TermCache(new TermDao(controller.getRecman()));
		this.familyProvider = new FamilyCache(new FamilyDao(
				controller.getRecman()));
		this.subjectProvider = new SubjectCache(new SubjectDao(
				controller.getRecman()));

		this.documentProvider = new DocumentElementDao(controller.getRecman());
		this.labeledDocumentProvider = new LabeledDocumentDao(controller.getRecman());
	}

	public Integer getTermId(String term) {
		return this.termProvider.getId(term);
	}

	/**
	 * @return o vetor de idf dos termos da coleção
	 */
	public double[] getIdf() {
		if (this.idf == null) {
			Term[] terms = this.termProvider.terms();
			double[] idf = new double[terms.length];

			for (int i = 0; i < idf.length; i++) {
				if (terms[i] != null) {
					idf[i] = Math.log1p(((double) this.getSize())
							/ ((double) (terms[i].getCount() + 1)));
				}

			}
			this.idf = idf;
		}
		return this.idf;

	}

	public void incrementTermFrequency(Integer termId) {
		this.termProvider.increment(termId);
	}

	/**
	 * 
	 * @return a quantidade de documentos da coleção
	 */
	public int getSize() {
		int size = this.documentProvider.size();
		return size;
	}

	public DocumentElement getDocument(Integer id) {
		return this.documentProvider.getDocumentElement(id);
	}

	public void saveDocumentElements(Collection<DocumentElement> elements) {
		for (DocumentElement element : elements) {
			this.saveDocumentElement(element);
		}
	}

	private void saveDocumentElement(DocumentElement element) {
		this.idf = null;
		this.documentProvider.save(element);
	}

	public void saveSubjects(Collection<Subject> subjects) {
		for (Subject subject : subjects) {
			this.subjectProvider.save(subject);
		}
	}

	public Iterable<Subject> getActiveSubjects() {
		return this.subjectProvider.subjects();
	}

	public Subject createSubject(DocumentElement element) {
		Subject subject = new Subject();
		subject.setValues(element.getIndexes(), element.getValues());
		subject.setFamilyId(1);

		this.subjectProvider.save(subject);
		element.setSubjectId(subject.getId());

		return subject;
	}

	public synchronized void queueAdd(DocumentElement element) {
		this.queue.add(element);
		this.saveDocumentElement(element);
	}

	public synchronized void queueAdd(List<DocumentElement> elements) {
		this.queue.addAll(elements);
		this.saveDocumentElements(elements);
	}

	public synchronized Collection<DocumentElement> queueValues() {
		Collection<DocumentElement> ret = this.queue;
		this.queue = new LinkedList<DocumentElement>();
		return ret;
	}

	public ClassificationResult categoryOf(Document document) {
		DocumentElement element = this.controller.justProcessDoc(document);
		this.saveDocumentElement(element);

		Subject subject = new Subject();
		subject.setValues(element.getIndexes(), element.getValues());

		return this.controller.getClassificator().categoryOf(subject, this);
	}

	public void setCategory(Integer documentId, Integer label) {
		DocumentElement element = this.documentProvider
				.getDocumentElement(documentId);

		LabeledDocument labdoc = new LabeledDocument(element, label);
		this.saveLabeledDocument(labdoc);

		this.controller.getClassificator().learn(element, label);

	}

	public void setCategory(Document document, Integer label) {
		DocumentElement element = this.controller.justProcessDoc(document);
		this.saveDocumentElement(element);

		LabeledDocument labdoc = new LabeledDocument(element, label);
		this.saveLabeledDocument(labdoc);

		this.controller.getClassificator().learn(element, label);
	}

	public void saveLabeledDocument(LabeledDocument doc) {
		this.labeledDocumentProvider.save(doc);
	}

	public java.util.Collection<LabeledDocument> getLabeledDocuments() {
		return this.labeledDocumentProvider.values();
	}

	public int getLabeledDocumentsSize() {
		return this.labeledDocumentProvider.size();
	}

	public void commit() throws IOException {
		Cronometro.COMMIT.start();
		this.controller.getClassificator().commit();
		this.controller.getClustering().commit();

		this.familyProvider.commit();
		this.termProvider.commit();
		this.subjectProvider.commit();

		this.controller.getRecman().commit();
		this.controller.getRecman().clearCache();
		Cronometro.COMMIT.stop();
	}

	public Subject getSubject(Integer id) {
		return this.subjectProvider.getSubject(id);
	}

	public void close() throws IOException {
		this.controller.getRecman().close();
	}

	public void removeSubject(int id) {
		this.subjectProvider.remove(id);
	}

	public DocumentProvider getDocumentProvider() {
		return this.documentProvider;
	}
}
