package bnrobot.core.parser;

import java.io.StringReader;
import java.text.Normalizer;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.TokenStream;
import org.apache.lucene.analysis.tokenattributes.TermAttribute;

import bnrobot.core.DocCollection;
import bnrobot.core.Document;

/**
 * Classe auxiliar, usada apenas para contagem dos termos de um documento.
 * 
 * @author Willian Gigliotti
 * 
 */
class TermCounter implements Comparable<TermCounter> {
	private int termId;
	private int count = 0;

	public TermCounter(int termId) {

		this.termId = termId;
	}

	public int compareTo(TermCounter arg0) {
		return (int) (this.termId - arg0.termId);
	}

	public void increase() {
		this.count++;
	}

	public int getTermId() {
		return this.termId;
	}

	public int getCount() {
		return this.count;
	}

}

/**
 * Classe usada para contar a frequencia dos termos num documento
 * 
 * @author willian
 * 
 */

public class DocumentParser {
	private Document doc;
	private DocCollection collection;
	private Analyzer analyzer;

	private int[] indexes;
	private int[] frequency;

	private static final Pattern tagPattern = Pattern.compile("<[^>]*>");
	private static final Pattern clrPattern = Pattern.compile("[^a-zA-Z_\\-]+");
	private static final Pattern nrmPattern = Pattern.compile("[^\\p{ASCII}]");

	/**
	 * 
	 * @param doc
	 *            Documento que será parseado
	 * @param collection
	 *            coleção em que o documento será analizado
	 * @param analyzer
	 *            Analyzer do Lucene, de preferência com lematizador, stop words
	 *            ;)
	 */
	public DocumentParser(Document doc, DocCollection collection,
			Analyzer analyzer) {
		this.analyzer = analyzer;
		this.doc = doc;
		this.collection = collection;
	}

	private String clearContent(String document) {
		String replace = " ";
		Matcher m;

		document = Normalizer.normalize(document, Normalizer.Form.NFD);
		// documento = documento.replaceAll(Term.nrmPattern, "");

		// Deixando todos os caracteres ASCII
		m = DocumentParser.nrmPattern.matcher(document);
		document = m.replaceAll("");

		// Tirando TAGs HTMLs do documento
		document = DocumentParser.tagPattern.matcher(document).replaceAll(
				replace);

		// Limpando caracteres estranhos
		document = DocumentParser.clrPattern.matcher(document).replaceAll(
				replace);

		// deixando letras minusculas
		document = document.toLowerCase();

		return document;
	}

	private Map<String, TermCounter> consumeDocument() {

		HashMap<String, TermCounter> terms = new HashMap<String, TermCounter>();

		StringReader sr = new StringReader(this.clearContent(this.doc
				.getContent()));
		TokenStream ts = this.analyzer.tokenStream(null, sr);

		try {
			ts.reset();

			while (ts.incrementToken()) {

				TermAttribute ta = ts.getAttribute(TermAttribute.class);
				String term = ta.term();

				if (!terms.containsKey(term)) {
					terms.put(term, new TermCounter(collection.getTermId(term)));

				}
				terms.get(term).increase();

			}
			ts.end();
			ts.close();

		} catch (Exception e) {
			// TODO: handle exception
		}

		return terms;

	}

	public void parse() {

		Map<String, TermCounter> terms = this.consumeDocument();

		LinkedList<TermCounter> list = new LinkedList<TermCounter>(
				terms.values());
		Collections.sort(list);

		this.indexes = new int[list.size()];
		this.frequency = new int[list.size()];

		int count = 0;
		for (TermCounter t : list) {
			this.indexes[count] = t.getTermId();
			this.frequency[count] = t.getCount();
			count++;
		}

	}

	public int[] getIndexes() {
		if (this.indexes == null) {
			this.parse();
		}
		return this.indexes;
	}

	public int[] getFrequency() {
		if (this.frequency == null) {
			this.parse();
		}
		return this.frequency;
	}

}
